package migrateservice

import (
	"context"
	"fmt"
	"net/http"
	"os"
	"time"

	"github.com/go-gorp/gorp"
	"github.com/gorilla/mux"

	"github.com/ovh/cds/engine/api"
	"github.com/ovh/cds/engine/api/database"
	"github.com/ovh/cds/engine/api/services"
	"github.com/ovh/cds/engine/service"
	"github.com/ovh/cds/sdk"
	"github.com/ovh/cds/sdk/cdsclient"
	"github.com/ovh/cds/sdk/log"
)

type dbmigservice struct {
	service.Common
	cfg           Configuration
	Router        *api.Router
	currentStatus struct {
		err        error
		migrations []sdk.DatabaseMigrationStatus
	}
}

var _ service.BeforeStart = new(dbmigservice)

// Configuration is the exposed type for database API configuration
type Configuration struct {
	Name      string `toml:"name" comment:"Name of this CDS Database Migrate service\n Enter a name to enable this service" json:"name"`
	URL       string `default:"http://localhost:8087" json:"url"`
	Directory string `toml:"directory" comment:"SQL Migration files directory" default:"sql" json:"directory"`
	HTTP      struct {
		Addr     string `toml:"addr" default:"" commented:"true" comment:"Listen address without port, example: 127.0.0.1" json:"addr"`
		Port     int    `toml:"port" default:"8087" json:"port"`
		Insecure bool   `toml:"insecure" default:"false" commented:"true" comment:"sslInsecureSkipVerify, set to true if you use a self-signed SSL on CDS API" json:"insecure"`
	} `toml:"http" comment:"######################\n CDS DB Migrate HTTP Configuration \n######################" json:"http"`
	API service.APIServiceConfiguration `toml:"api" comment:"######################\n CDS API Settings \n######################" json:"api"`
	DB  database.DBConfiguration        `toml:"db" comment:"################################\n Postgresql Database settings \n###############################" json:"db"`
}

// New instanciates a new API object
func New() service.Service {
	return &dbmigservice{}
}

func (s *dbmigservice) Init(config interface{}) (cdsclient.ServiceConfig, error) {
	var cfg cdsclient.ServiceConfig
	sConfig, ok := config.(Configuration)
	if !ok {
		return cfg, sdk.WithStack(fmt.Errorf("invalid migrate service configuration"))
	}

	cfg.Host = sConfig.API.HTTP.URL
	cfg.Token = sConfig.API.Token
	cfg.InsecureSkipVerifyTLS = sConfig.API.HTTP.Insecure
	cfg.RequestSecondsTimeout = sConfig.API.RequestTimeout
	return cfg, nil
}

func (s *dbmigservice) CheckConfiguration(cfg interface{}) error {
	_, ok := cfg.(Configuration)
	if !ok {
		return fmt.Errorf("Invalid configuration")
	}
	return nil
}

func (s *dbmigservice) ApplyConfiguration(cfg interface{}) error {
	if err := s.CheckConfiguration(cfg); err != nil {
		return err
	}

	dbCfg, _ := cfg.(Configuration)

	s.cfg = dbCfg
	s.ServiceName = s.cfg.Name
	s.ServiceType = services.TypeDBMigrate
	s.HTTPURL = s.cfg.URL

	s.MaxHeartbeatFailures = s.cfg.API.MaxHeartbeatFailures
	s.Router = &api.Router{
		Mux: mux.NewRouter(),
	}
	return nil
}

func (s *dbmigservice) BeforeStart(ctx context.Context) error {
	log.Info(ctx, "DBMigrate> Starting Database migration...")
	dbConn, err := database.Init(
		ctx,
		s.cfg.DB.User,
		s.cfg.DB.Role,
		s.cfg.DB.Password,
		s.cfg.DB.Name,
		s.cfg.DB.Host,
		s.cfg.DB.Port,
		s.cfg.DB.SSLMode,
		s.cfg.DB.ConnectTimeout,
		s.cfg.DB.Timeout,
		s.cfg.DB.MaxConn)
	if err != nil {
		return sdk.WrapError(fmt.Errorf("cannot connect to database: %v", err), "getMigrate")
	}

	upgradeTo := os.Getenv("UPGRADE_TO")     // Set this env variable to define the maximum migration file you want to upgrade
	downgradeTo := os.Getenv("DOWNGRADE_TO") // Set this env variable to define the maximum migration file you want to downgrade

	errDo := s.doMigrate(dbConn.DB, gorp.PostgresDialect{}, upgradeTo, downgradeTo)
	if errDo != nil {
		log.Error(ctx, "DBMigrate> Migration failed %v", errDo)
		s.currentStatus.err = errDo
	}

	log.Info(ctx, "DBMigrate> Retrieving Database migration status...")
	status, errGet := s.getMigrate(dbConn.DB, gorp.PostgresDialect{})
	if errGet != nil {
		log.Error(ctx, "DBMigrate> Migration status unavailable %v", errGet)
	}
	if errDo == nil && errGet != nil {
		s.currentStatus.err = errGet
	}
	s.currentStatus.migrations = status

	// From now the database access won't be used. Erase the configuration...
	// This limits the attack surface
	s.cfg.DB = database.DBConfiguration{}

	return nil
}

func (s *dbmigservice) Serve(ctx context.Context) error {
	log.Info(ctx, "DBMigrate> Starting service %s %s...", s.cfg.Name, sdk.VERSION)
	s.StartupTime = time.Now()

	//Init the http server
	s.initRouter(ctx)
	server := &http.Server{
		Addr:           fmt.Sprintf("%s:%d", s.cfg.HTTP.Addr, s.cfg.HTTP.Port),
		Handler:        s.Router.Mux,
		ReadTimeout:    10 * time.Minute,
		WriteTimeout:   10 * time.Minute,
		MaxHeaderBytes: 1 << 20,
	}

	go func() {
		//Start the http server
		log.Info(ctx, "DBMigrate> Starting HTTP Server on port %d", s.cfg.HTTP.Port)
		if err := server.ListenAndServe(); err != nil {
			log.Error(ctx, "DBMigrate> Listen and serve failed: %s", err)
		}
	}()

	//Gracefully shutdown the http server
	<-ctx.Done()
	log.Info(ctx, "DBMigrate> Shutdown HTTP Server")
	if err := server.Shutdown(ctx); err != nil {
		return fmt.Errorf("unable to shutdown server: %v", err)
	}

	return ctx.Err()
}

func (s *dbmigservice) Status(ctx context.Context) sdk.MonitoringStatus {
	response := s.CommonMonitoring()
	if s.currentStatus.err != nil {
		response.Lines = append(response.Lines,
			sdk.MonitoringStatusLine{
				Component: "SQL",
				Value:     s.currentStatus.err.Error(),
				Status:    sdk.MonitoringStatusAlert,
			},
		)
		return response
	}

	var theNumberOfSuccessfulMigations int
	for _, m := range s.currentStatus.migrations {
		if m.Migrated {
			theNumberOfSuccessfulMigations++
		}
	}

	var status = sdk.MonitoringStatusWarn
	if theNumberOfSuccessfulMigations == len(s.currentStatus.migrations) {
		status = sdk.MonitoringStatusOK
	}

	response.Lines = append(response.Lines,
		sdk.MonitoringStatusLine{
			Component: "SQL",
			Value:     fmt.Sprintf("%d/%d", theNumberOfSuccessfulMigations, len(s.currentStatus.migrations)),
			Status:    status,
		},
	)

	return response
}

func (s *dbmigservice) initRouter(ctx context.Context) {
	log.Debug("DBMigrate> Router initialized")
	r := s.Router
	r.SetHeaderFunc = api.DefaultHeaders
	r.Middlewares = append(r.Middlewares, service.CheckRequestSignatureMiddleware(s.ParsedAPIPublicKey))

	r.Handle("/mon/version", nil, r.GET(api.VersionHandler, api.Auth(false)))
	r.Handle("/mon/status", nil, r.GET(s.statusHandler, api.Auth(false)))
	r.Handle("/mon/metrics", nil, r.GET(service.GetPrometheustMetricsHandler(s), api.Auth(false)))
	r.Handle("/mon/metrics/all", nil, r.GET(service.GetMetricsHandler, api.Auth(false)))
	r.Handle("/", nil, r.GET(s.getMigrationHandler, api.Auth(false)))
}
